# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Callable, Tuple, List
from dataclasses import dataclass
import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers.activations import ACT2FN
from transformers import AutoConfig
import cosmos_rl.policy.kernel.modeling_utils as modeling_utils


@dataclass
class Qwen3VLMoe_Encoder_Args:
    depth: int
    hidden_size: int
    hidden_act: str
    intermediate_size: int
    n_heads: int
    in_channels: int
    patch_size: int
    spatial_merge_size: int
    temporal_patch_size: int
    num_position_embeddings: int
    out_hidden_size: int
    deepstack_visual_indexes: List[int]
    norm_type: str = "layernorm"
    layer_norm_eps: float = 1e-6
    hf_config: AutoConfig = None


class Qwen3VLMoeVisionMLP(nn.Module):
    def __init__(self, config: Qwen3VLMoe_Encoder_Args, bias: bool = True):
        super().__init__()
        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size
        self.linear_fc1 = nn.Linear(self.hidden_size, self.intermediate_size, bias=bias)
        self.linear_fc2 = nn.Linear(self.intermediate_size, self.hidden_size, bias=bias)
        self.act_fn = ACT2FN[config.hidden_act]

    def forward(self, hidden_state):
        return self.linear_fc2(self.act_fn(self.linear_fc1(hidden_state)))


class Qwen3VLMoeVisionPatchEmbed(nn.Module):
    def __init__(
        self,
        patch_size: int = 16,
        temporal_patch_size: int = 2,
        in_channels: int = 3,
        embed_dim: int = 1152,
    ) -> None:
        super().__init__()
        self.patch_size = patch_size
        self.temporal_patch_size = temporal_patch_size
        self.in_channels = in_channels
        self.embed_dim = embed_dim

        kernel_size = [temporal_patch_size, patch_size, patch_size]
        self.proj = nn.Conv3d(
            in_channels,
            embed_dim,
            kernel_size=kernel_size,
            stride=kernel_size,
            bias=True,
        )

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        target_dtype = self.proj.weight.dtype
        hidden_states = hidden_states.view(
            -1,
            self.in_channels,
            self.temporal_patch_size,
            self.patch_size,
            self.patch_size,
        )
        hidden_states = self.proj(hidden_states.to(dtype=target_dtype)).view(
            -1, self.embed_dim
        )
        return hidden_states


class Qwen3VLMoeVisionRotaryEmbedding(nn.Module):
    inv_freq: torch.Tensor  # fix linting for `register_buffer`

    def __init__(self, dim: int, theta: float = 10000.0) -> None:
        super().__init__()
        self.dim = dim
        self.theta = theta
        self.reset_inv_freq()

    def reset_inv_freq(self, device: torch.device = None):
        inv_freq = 1.0 / (
            self.theta ** (torch.arange(0, self.dim, 2, dtype=torch.float) / self.dim)
        )
        inv_freq = inv_freq.to(device) if device is not None else inv_freq
        if not hasattr(self, "inv_freq"):
            self.register_buffer("inv_freq", inv_freq, persistent=False)
        else:
            self.inv_freq.to(torch.float32)
            with torch.no_grad():
                self.inv_freq.data.copy_(inv_freq)

    def forward(self, seqlen: int, device: torch.device = None) -> torch.Tensor:
        if self.inv_freq.dtype != torch.float32:
            self.reset_inv_freq(device=device)
            assert (
                self.inv_freq.dtype == torch.float32
            ), "inv_freq dtype should be float32"
        seq = torch.arange(
            seqlen, device=self.inv_freq.device, dtype=self.inv_freq.dtype
        )
        freqs = torch.outer(seq, self.inv_freq)
        return freqs


class Qwen3VLMoeVisionPatchMerger(nn.Module):
    def __init__(
        self, config: Qwen3VLMoe_Encoder_Args, use_postshuffle_norm=False
    ) -> None:
        super().__init__()
        self.hidden_size = config.hidden_size * (config.spatial_merge_size**2)
        self.use_postshuffle_norm = use_postshuffle_norm
        self.norm = nn.LayerNorm(
            self.hidden_size if use_postshuffle_norm else config.hidden_size, eps=1e-6
        )
        self.linear_fc1 = nn.Linear(self.hidden_size, self.hidden_size)
        self.act_fn = nn.GELU()
        self.linear_fc2 = nn.Linear(self.hidden_size, config.out_hidden_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm(
            x.view(-1, self.hidden_size) if self.use_postshuffle_norm else x
        ).view(-1, self.hidden_size)
        x = self.linear_fc2(self.act_fn(self.linear_fc1(x)))
        return x


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


if os.environ.get("COSMOS_USE_HF_IMPL", "0").lower() in ["1", "true"]:

    def apply_rotary_pos_emb_vision(
        q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        orig_q_dtype = q.dtype
        orig_k_dtype = k.dtype
        q, k = q.float(), k.float()
        cos, sin = cos.unsqueeze(-2).float(), sin.unsqueeze(-2).float()
        q_embed = (q * cos) + (rotate_half(q) * sin)
        k_embed = (k * cos) + (rotate_half(k) * sin)
        q_embed = q_embed.to(orig_q_dtype)
        k_embed = k_embed.to(orig_k_dtype)
        return q_embed, k_embed
else:

    def apply_rotary_pos_emb_vision(
        q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Apply rotary position embedding to the query and key tensors.
        """
        cos = cos.chunk(2, dim=-1)[0].contiguous()
        sin = sin.chunk(2, dim=-1)[0].contiguous()
        q_embed = modeling_utils.apply_rotary_emb(
            q.float(), cos.float(), sin.float()
        ).type_as(q)
        k_embed = modeling_utils.apply_rotary_emb(
            k.float(), cos.float(), sin.float()
        ).type_as(k)
        return q_embed, k_embed


class Qwen3VLMoeVisionAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int = 16) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.qkv = nn.Linear(dim, dim * 3, bias=True)
        self.proj = nn.Linear(dim, dim)
        self.attention_dropout = 0.0
        self.attn_func = modeling_utils.flash_attn_varlen_func

    def forward(
        self,
        hidden_states: torch.Tensor,
        cu_seqlens: torch.Tensor,
        position_embeddings: torch.Tensor,
        max_seqlen: int,
    ) -> torch.Tensor:
        seq_length = hidden_states.shape[0]
        q, k, v = (
            self.qkv(hidden_states)
            .reshape(seq_length, 3, self.num_heads, -1)
            .permute(1, 0, 2, 3)
            .unbind(0)
        )
        cos, sin = position_embeddings

        q, k = apply_rotary_pos_emb_vision(q.unsqueeze(0), k.unsqueeze(0), cos, sin)
        q = q.squeeze(0)
        k = k.squeeze(0)

        input_dtype = q.dtype
        if input_dtype == torch.float32:
            if torch.is_autocast_enabled():
                target_dtype = torch.get_autocast_gpu_dtype()
            else:
                raise ValueError("Flash attention only supports float32 input")
            q = q.to(target_dtype)
            k = k.to(target_dtype)
            v = v.to(target_dtype)

        attn_output = self.attn_func(
            q,
            k,
            v,
            cu_seqlens_q=cu_seqlens,
            cu_seqlens_k=cu_seqlens,
            max_seqlen_q=max_seqlen,
            max_seqlen_k=max_seqlen,
            causal=False,
        ).reshape(seq_length, -1)
        attn_output = self.proj(attn_output)
        return attn_output


class Qwen3VLMoeVisionBlock(nn.Module):
    def __init__(self, config: Qwen3VLMoe_Encoder_Args) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(config.hidden_size, eps=1e-6)
        self.norm2 = nn.LayerNorm(config.hidden_size, eps=1e-6)
        self.attn = Qwen3VLMoeVisionAttention(
            config.hidden_size, num_heads=config.n_heads
        )
        self.mlp = Qwen3VLMoeVisionMLP(config=config, bias=True)

    def forward(
        self, hidden_states, cu_seqlens, position_embeddings, max_seqlen
    ) -> torch.Tensor:
        hidden_states = hidden_states + self.attn(
            self.norm1(hidden_states),
            cu_seqlens=cu_seqlens,
            position_embeddings=position_embeddings,
            max_seqlen=max_seqlen,
        )
        hidden_states = hidden_states + self.mlp(self.norm2(hidden_states))
        return hidden_states


class Qwen3VLMoeVisionModel(nn.Module):
    def __init__(self, config) -> None:
        super().__init__()
        self.config = config
        self.spatial_merge_size = config.spatial_merge_size
        self.patch_size = config.patch_size
        self.spatial_merge_unit = self.spatial_merge_size * self.spatial_merge_size

        self.patch_embed = Qwen3VLMoeVisionPatchEmbed(
            patch_size=config.patch_size,
            temporal_patch_size=config.temporal_patch_size,
            in_channels=config.in_channels,
            embed_dim=config.hidden_size,
        )

        self.pos_embed = nn.Embedding(
            config.num_position_embeddings, config.hidden_size
        )
        self.num_grid_per_side = int(config.num_position_embeddings**0.5)

        head_dim = config.hidden_size // config.n_heads
        self.rotary_pos_emb = Qwen3VLMoeVisionRotaryEmbedding(head_dim // 2)

        self.blocks = torch.nn.ModuleDict()
        for layer_id in range(config.depth):
            self.blocks[str(layer_id)] = Qwen3VLMoeVisionBlock(config)

        self.merger = Qwen3VLMoeVisionPatchMerger(
            config=config,
            use_postshuffle_norm=False,
        )

        self.deepstack_visual_indexes = config.deepstack_visual_indexes
        self.deepstack_merger_list = nn.ModuleList(
            [
                Qwen3VLMoeVisionPatchMerger(
                    config=config,
                    use_postshuffle_norm=True,
                )
                for _ in range(len(config.deepstack_visual_indexes))
            ]
        )

        self.gradient_checkpointing = False

    def rot_pos_emb(self, grid_thw: torch.Tensor) -> torch.Tensor:
        merge_size = self.spatial_merge_size

        max_hw = int(grid_thw[:, 1:].max().item())
        freq_table = self.rotary_pos_emb(max_hw)  # (max_hw, dim // 2)
        device = freq_table.device

        total_tokens = int(torch.prod(grid_thw, dim=1).sum().item())
        pos_ids = torch.empty((total_tokens, 2), dtype=torch.long, device=device)

        offset = 0
        for num_frames, height, width in grid_thw:
            merged_h, merged_w = height // merge_size, width // merge_size

            block_rows = torch.arange(merged_h, device=device)  # block row indices
            block_cols = torch.arange(merged_w, device=device)  # block col indices
            intra_row = torch.arange(
                merge_size, device=device
            )  # intra-block row offsets
            intra_col = torch.arange(
                merge_size, device=device
            )  # intra-block col offsets

            # Compute full-resolution positions
            row_idx = (
                block_rows[:, None, None, None] * merge_size
                + intra_row[None, None, :, None]
            )
            col_idx = (
                block_cols[None, :, None, None] * merge_size
                + intra_col[None, None, None, :]
            )

            row_idx = row_idx.expand(
                merged_h, merged_w, merge_size, merge_size
            ).reshape(-1)
            col_idx = col_idx.expand(
                merged_h, merged_w, merge_size, merge_size
            ).reshape(-1)

            coords = torch.stack((row_idx, col_idx), dim=-1)

            if num_frames > 1:
                coords = coords.repeat(num_frames, 1)

            num_tokens = coords.shape[0]
            pos_ids[offset : offset + num_tokens] = coords
            offset += num_tokens

        embeddings = freq_table[pos_ids]  # lookup rotary embeddings
        embeddings = embeddings.flatten(1)
        return embeddings

    def fast_pos_embed_interpolate(self, grid_thw):
        grid_ts, grid_hs, grid_ws = grid_thw[:, 0], grid_thw[:, 1], grid_thw[:, 2]

        idx_list = [[] for _ in range(4)]
        weight_list = [[] for _ in range(4)]

        for t, h, w in zip(grid_ts, grid_hs, grid_ws):
            h_idxs = torch.linspace(0, self.num_grid_per_side - 1, h)
            w_idxs = torch.linspace(0, self.num_grid_per_side - 1, w)

            h_idxs_floor = h_idxs.int()
            w_idxs_floor = w_idxs.int()
            h_idxs_ceil = (h_idxs.int() + 1).clip(max=self.num_grid_per_side - 1)
            w_idxs_ceil = (w_idxs.int() + 1).clip(max=self.num_grid_per_side - 1)

            dh = h_idxs - h_idxs_floor
            dw = w_idxs - w_idxs_floor

            base_h = h_idxs_floor * self.num_grid_per_side
            base_h_ceil = h_idxs_ceil * self.num_grid_per_side

            indices = [
                (base_h[None].T + w_idxs_floor[None]).flatten(),
                (base_h[None].T + w_idxs_ceil[None]).flatten(),
                (base_h_ceil[None].T + w_idxs_floor[None]).flatten(),
                (base_h_ceil[None].T + w_idxs_ceil[None]).flatten(),
            ]

            weights = [
                ((1 - dh)[None].T * (1 - dw)[None]).flatten(),
                ((1 - dh)[None].T * dw[None]).flatten(),
                (dh[None].T * (1 - dw)[None]).flatten(),
                (dh[None].T * dw[None]).flatten(),
            ]

            for i in range(4):
                idx_list[i].extend(indices[i].tolist())
                weight_list[i].extend(weights[i].tolist())

        idx_tensor = torch.tensor(
            idx_list, dtype=torch.long, device=self.pos_embed.weight.device
        )
        weight_tensor = torch.tensor(
            weight_list,
            dtype=self.pos_embed.weight.dtype,
            device=self.pos_embed.weight.device,
        )
        pos_embeds = self.pos_embed(idx_tensor) * weight_tensor[:, :, None]
        patch_pos_embeds = pos_embeds[0] + pos_embeds[1] + pos_embeds[2] + pos_embeds[3]

        patch_pos_embeds = patch_pos_embeds.split(
            [h * w for h, w in zip(grid_hs, grid_ws)]
        )

        patch_pos_embeds_permute = []
        merge_size = self.config.spatial_merge_size
        for pos_embed, t, h, w in zip(patch_pos_embeds, grid_ts, grid_hs, grid_ws):
            pos_embed = pos_embed.repeat(t, 1)
            pos_embed = (
                pos_embed.view(
                    t, h // merge_size, merge_size, w // merge_size, merge_size, -1
                )
                .permute(0, 1, 3, 2, 4, 5)
                .flatten(0, 4)
            )
            patch_pos_embeds_permute.append(pos_embed)
        patch_pos_embeds = torch.cat(patch_pos_embeds_permute)
        return patch_pos_embeds

    def forward(
        self, hidden_states: torch.Tensor, grid_thw: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """
        Args:
            hidden_states (`torch.Tensor` of shape `(seq_len, hidden_size)`):
                The final hidden states of the model.
            grid_thw (`torch.Tensor` of shape `(num_images_or_videos, 3)`):
                The temporal, height and width of feature shape of each image in LLM.

        Returns:
            `torch.Tensor`: hidden_states.
        """
        hidden_states = self.patch_embed(hidden_states)

        pos_embeds = self.fast_pos_embed_interpolate(grid_thw)
        hidden_states = hidden_states + pos_embeds

        rotary_pos_emb = self.rot_pos_emb(grid_thw)

        seq_len, _ = hidden_states.size()
        hidden_states = hidden_states.reshape(seq_len, -1)
        rotary_pos_emb = rotary_pos_emb.reshape(seq_len, -1)
        emb = torch.cat((rotary_pos_emb, rotary_pos_emb), dim=-1)
        position_embeddings = (emb.cos(), emb.sin())

        cu_seqlens = torch.repeat_interleave(
            grid_thw[:, 1] * grid_thw[:, 2], grid_thw[:, 0]
        ).cumsum(
            dim=0,
            # Select dtype based on the following factors:
            #  - FA2 requires that cu_seqlens_q must have dtype int32
            #  - torch.onnx.export requires that cu_seqlens_q must have same dtype as grid_thw
            # See https://github.com/huggingface/transformers/pull/34852 for more information
            dtype=grid_thw.dtype if torch.jit.is_tracing() else torch.int32,
        )
        cu_seqlens = F.pad(cu_seqlens, (1, 0), value=0)
        with torch.no_grad():
            max_seqlen = (cu_seqlens[1:] - cu_seqlens[:-1]).max().item()

        deepstack_feature_lists = []
        for layer_num, blk in self.blocks.items():
            if (
                hasattr(blk, "_gradient_checkpointing_enabled")
                and blk._gradient_checkpointing_enabled
            ):
                hidden_states = torch.utils.checkpoint.checkpoint(
                    blk,
                    hidden_states,
                    cu_seqlens,
                    position_embeddings,
                    max_seqlen,
                    use_reentrant=False,
                )
            else:
                hidden_states = blk(
                    hidden_states,
                    cu_seqlens=cu_seqlens,
                    position_embeddings=position_embeddings,
                    max_seqlen=max_seqlen,
                )
            if int(layer_num) in self.deepstack_visual_indexes:
                deepstack_feature = self.deepstack_merger_list[
                    self.deepstack_visual_indexes.index(int(layer_num))
                ](hidden_states)
                deepstack_feature_lists.append(deepstack_feature)

        hidden_states = self.merger(hidden_states)

        return hidden_states, deepstack_feature_lists

    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.blocks),
            self.config.n_heads,
            self.config.hidden_size // self.config.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import torch
import subprocess
import atexit
import sys
import uuid
import asyncio
import time
import os
import math
import threading
import tempfile
from typing import List, Dict, Tuple, Any, Optional, Callable
from cosmos_rl.dispatcher.replica import Atom, Rollout
from cosmos_rl.dispatcher.protocol import Role, MESH_NAMES
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.wandb_logger import (
    is_wandb_available,
    init_wandb,
)
import cosmos_rl.utils.util as util
import cosmos_rl.utils.network_util as network_util
import cosmos_rl.utils.constant as constant
from cosmos_rl.dispatcher.data import (
    CosmosDataset,
    RLPayload,
    CosmosValidationDataset,
)
from torch.utils.data import DataLoader, Dataset, DistributedSampler
from cosmos_rl.utils.redis_stream import RedisStreamHandler
from cosmos_rl.dispatcher.status import (
    PolicyStatusManager,
    RolloutStatusManager,
)
from cosmos_rl.policy.config import Config, SubProfilerConfig
from cosmos_rl.dispatcher.protocol import SetProfileRequest
from transformers import AutoTokenizer
from cosmos_rl.dispatcher.data.packer.base import DataPacker
from cosmos_rl.dispatcher.command import PolicyToRolloutUnicastCommand
from cosmos_rl.utils.checkpoint import CheckpointMananger
from cosmos_rl.utils.parallelism_map import ParallelizedShardMapper
from cosmos_rl.dispatcher.data import IdxAndRLPayload
from concurrent.futures import ProcessPoolExecutor
from itertools import islice


class Controller:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Controller, cls).__new__(cls)
            cls._instance._init_dist()
        return cls._instance

    def __init__(self):
        if not hasattr(self, "config"):
            self._init_dist()
        self._init_status()

    def _init_status(self):
        self.policy_status_manager = PolicyStatusManager()
        self.rollout_status_manager = RolloutStatusManager()
        self.epoch = 1
        self.stat_prompt_tokens_count = 0
        self.stat_completion_tokens_count = 0
        self.stat_n_samples = 0
        self.begin_time = None
        # nccl error check
        self.post_ncclerror_policy_invoke_id = 0
        self.post_ncclerror_rollout_invoke_id = 0

    def _init_dist(self):
        self.config = None
        self.temp_kv_store = {}

        self.life_cycle_lock = asyncio.Lock()
        self.shut_down_event = threading.Event()

    def setup(
        self,
        config: Config,
        redis_port: int,
        redis_logfile_path: str,
        dataset: Optional[Dataset] = None,
        data_packer: Optional[DataPacker] = None,
        val_dataset: Optional[Dataset] = None,
        val_data_packer: Optional[DataPacker] = None,
        custom_logger_fns: Optional[List[Callable]] = None,
        sampler: Optional[Callable] = None,
        batch_sampler: Optional[Callable] = None,
        val_sampler: Optional[Callable] = None,
        val_batch_sampler: Optional[Callable] = None,
    ):
        if self.config is not None:
            raise Exception(
                "[Controller] Config has been set. Please do not call setup again."
            )

        self.config = config
        task_type = config.train.train_policy.type
        self.tokenizer = util.retry(AutoTokenizer.from_pretrained)(
            config.policy.model_name_or_path,
            trust_remote_code=True,
        )
        self.policy_to_rollout_shard_mapper = ParallelizedShardMapper.get_instance(
            config
        )

        if "wandb" in config.logging.logger and is_wandb_available():
            init_wandb(config)
        else:
            logger.warning(
                "Wandb is not available. Please install it to use wandb logging features."
            )

        self.is_rl = task_type != "sft"

        if dataset is not None and isinstance(dataset, Callable):
            dataset = dataset(config)
        if val_dataset is not None and isinstance(val_dataset, Callable):
            val_dataset = val_dataset(config)

        self.user_data_packer = data_packer
        self.user_val_data_packer = val_data_packer
        self.dataset = None
        self.ckpt_extra_info = {}
        self.weight_version_to_prompt_num = {}  # Only for on-policy.
        remain_samples_num = 0

        if self.is_rl:
            self.rollout_batch_size = (
                config.train.train_policy.dataloader_batch_size
                or config.rollout.batch_size
            )
            if dataset is not None:
                assert isinstance(dataset, Dataset)
                self.dataset = CosmosDataset(
                    config=config, train_set=dataset, tokenizer=self.tokenizer
                )
                logger.info(
                    "[Controller] Using provided dataset for training, dataset specification in the toml config will be ignored"
                )
            else:
                self.dataset = CosmosDataset(config=config, tokenizer=self.tokenizer)

            remain_samples_num = (
                (
                    len(self.dataset.train_set)
                    * config.rollout.n_generation
                    * config.train.epoch
                )
                if self.dataset is not None
                else 0
            )  # Total number of samples of policy training will consume.

            if sampler is not None:
                logger.info("[Controller] Using provided sampler for training")
                if isinstance(sampler, Callable):
                    train_sampler = sampler(
                        self.dataset.train_set,
                        num_replicas=1,
                        rank=0,
                        shuffle=config.train.train_policy.dataloader_shuffle,
                        drop_last=False,
                    )
                else:
                    train_sampler = sampler
            else:
                train_sampler = DistributedSampler(
                    self.dataset.train_set,
                    num_replicas=1,
                    rank=0,
                    shuffle=config.train.train_policy.dataloader_shuffle,
                    drop_last=False,
                )
            if batch_sampler is not None and isinstance(batch_sampler, Callable):
                batch_sampler = batch_sampler(
                    train_sampler,
                    batch_size=self.rollout_batch_size,
                    drop_last=False,
                )

            if config.train.resume:
                try:
                    # If resuming, disable the weight sync check flag for rollout to compare the received weight with the reference weight.
                    PolicyToRolloutUnicastCommand._do_weight_sync_check_flag = False
                    self.ckpt_manager = CheckpointMananger(config)
                    self.ckpt_extra_info = (
                        self.ckpt_manager.load_extra_info_from_checkpoint()
                    )
                    remain_samples_num = self.ckpt_extra_info.get(
                        "remain_samples_num", remain_samples_num
                    )
                    self.epoch = (
                        config.train.epoch
                        - (
                            math.ceil(
                                remain_samples_num
                                / (
                                    len(self.dataset.train_set)
                                    * config.rollout.n_generation
                                )
                            )
                        )
                        + 1
                    )
                    logger.info(
                        f"[Controller] Resuming from checkpoint, current epoch: {self.epoch}, remaining samples: {remain_samples_num}"
                    )

                    train_dataloader_bias = max(
                        0,
                        len(self.dataset.train_set)
                        - (
                            (
                                math.ceil(
                                    remain_samples_num / config.rollout.n_generation
                                )
                            )
                            % len(self.dataset.train_set)
                        ),
                    )
                    logger.info(
                        f"[Controller] Loaded extra info from checkpoint: {self.ckpt_extra_info}"
                    )
                    from cosmos_rl.policy.trainer.sampler import SkippingSampler

                    train_sampler = SkippingSampler(
                        base_sampler=train_sampler,
                        skip_samples=train_dataloader_bias
                        // (
                            len(list(islice(iter(train_sampler), 1))[0])
                            if isinstance(list(islice(iter(train_sampler), 1))[0], list)
                            else 1
                        ),
                    )

                    if batch_sampler is not None:
                        batch_sampler = SkippingSampler(
                            base_sampler=batch_sampler,
                            skip_samples=train_dataloader_bias
                            // (
                                len(list(islice(iter(batch_sampler), 1))[0])
                                if isinstance(
                                    list(islice(iter(batch_sampler), 1))[0], list
                                )
                                else 1
                            ),
                        )
                except Exception as e:
                    import traceback

                    traceback.print_exc()
                    logger.error(
                        f"[Controller] Failed to load checkpoint extra info: {e}. Please check the checkpoint path and config."
                    )
            if batch_sampler is not None:
                logger.info(
                    "[Controller] Using custom batch Sampler that yields list of indices for training dataset."
                )
                self.train_dataloader = DataLoader(
                    self.dataset.train_set,
                    num_workers=config.train.train_policy.dataloader_num_workers,
                    prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                    collate_fn=RLPayload.collate_fn,
                    batch_sampler=batch_sampler,
                )
            else:
                self.train_dataloader = DataLoader(
                    self.dataset.train_set,
                    batch_size=self.rollout_batch_size,
                    shuffle=False,
                    num_workers=config.train.train_policy.dataloader_num_workers,
                    prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                    collate_fn=RLPayload.collate_fn,
                    sampler=train_sampler,
                )
            self.train_dataloader_iter = iter(self.train_dataloader)

            if config.validation.enable:
                self.val_batch_size = (
                    config.train.train_policy.dataloader_batch_size
                    or config.validation.batch_size
                    or self.rollout_batch_size
                )
                assert (
                    self.val_batch_size > 0
                ), "[Controller] val_batch_size should be greater than 0."
                if val_dataset is not None:
                    assert isinstance(val_dataset, Dataset)
                    self.val_dataset = CosmosValidationDataset(
                        config=config, val_set=val_dataset, tokenizer=self.tokenizer
                    )
                    logger.info(
                        "[Controller] Using provided validation dataset for validation, dataset specification in the toml config will be ignored"
                    )
                else:
                    self.val_dataset = CosmosValidationDataset(
                        config=config, tokenizer=self.tokenizer
                    )
                if val_sampler is not None:
                    logger.info("[Controller] Using provided sampler for validation")
                    if isinstance(val_sampler, Callable):
                        val_sampler = val_sampler(
                            self.val_dataset.val_set,
                            num_replicas=1,
                            rank=0,
                            shuffle=False,
                            drop_last=False,
                        )
                if val_batch_sampler is not None:
                    logger.info(
                        "Using custom batch Sampler that yields list of indices for validation dataset."
                    )
                    if isinstance(val_batch_sampler, Callable):
                        val_batch_sampler = val_batch_sampler(
                            val_sampler
                            if val_sampler is not None
                            else DistributedSampler(
                                self.val_dataset.val_set,
                                num_replicas=1,
                                rank=0,
                                shuffle=False,
                                drop_last=False,
                            ),
                            batch_size=self.val_batch_size,
                            drop_last=False,
                        )
                    val_dataloader = DataLoader(
                        self.val_dataset.val_set,
                        num_workers=config.train.train_policy.dataloader_num_workers,
                        prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                        collate_fn=RLPayload.collate_fn,
                        batch_sampler=val_batch_sampler,
                    )
                else:
                    val_dataloader = DataLoader(
                        self.val_dataset.val_set,
                        batch_size=self.val_batch_size,
                        shuffle=False,
                        num_workers=config.train.train_policy.dataloader_num_workers,
                        prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                        collate_fn=RLPayload.collate_fn,
                        sampler=val_sampler,
                    )
            else:
                self.val_dataset = None
                val_dataloader = None
        else:
            self.val_dataset = None
            val_dataloader = None

        redis_free_port = util.find_available_port(redis_port)
        self.config.redis = str(redis_free_port)

        ips = network_util.get_eth_ips()
        if len(ips) > 0:
            self.config.eth_ips = ";".join(ips)

        random_db_file_name = f"cosmos_rl_{str(uuid.uuid4())}.rdb"
        config_file_path = tempfile.NamedTemporaryFile(
            delete=False, suffix=".redis_config.conf"
        )
        redis_cfg_path = util.write_redis_config(
            redis_free_port, redis_logfile_path, file_path=config_file_path.name
        )
        redis_server_cmd = f'redis-server {redis_cfg_path} --dbfilename {random_db_file_name} --save ""'

        redis_server_proc = subprocess.Popen(
            redis_server_cmd, shell=True, stdout=sys.stdout, stderr=sys.stderr
        )

        # Check if the redis server started successfully
        redis_server_proc.wait()
        ret_code = redis_server_proc.returncode

        if ret_code is not None and ret_code != 0:
            raise RuntimeError(
                f"Failed to start redis server with command: {redis_server_cmd} with return code {ret_code}"
            )
        else:
            logger.info(
                f"[Controller] Redis server started on port {redis_free_port} with command {redis_server_cmd}"
            )

        self.redis_controller = RedisStreamHandler(
            ips=["0.0.0.0"], port=redis_free_port
        )

        self.policy_status_manager.setup(
            config,
            self.redis_controller,
            remain_samples_num=remain_samples_num,
            samples_per_epoch=len(self.dataset.train_set) * config.rollout.n_generation
            if self.is_rl
            else 0,
            tokenizer=self.tokenizer,
            val_dataloader=val_dataloader,
            current_step=self.ckpt_extra_info.get("step", 0),
            max_num_steps=config.train.max_num_steps,
            custom_logger_fns=custom_logger_fns,
            val_datasize=len(self.val_dataset.val_set)
            if self.val_dataset is not None
            else 0,
        )
        self.rollout_status_manager.setup(
            config, self.redis_controller, tokenizer=self.tokenizer
        )

        self.reward_executor = ProcessPoolExecutor(max_workers=4)

        # Register the exit function to be called when the program exits
        def exit_server(redis_server_proc, redis_free_port):
            logger.info("Stopping redis server")
            redis_server_proc.terminate()
            redis_server_proc.wait()

            redis_terminate_cmd = f"redis-cli -p {redis_free_port} shutdown nosave"
            redis_terminate = subprocess.Popen(
                redis_terminate_cmd,
                shell=True,
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
            redis_terminate.wait()
            try:
                os.unlink(config_file_path.name)
            except Exception:
                # best effort to remove the config file
                pass
            logger.info("Redis server stopped.")

        atexit.register(exit_server, redis_server_proc, redis_free_port)

    async def update_kv_store(self, key: str, value: str):
        self.temp_kv_store[key] = value

    async def clear_temp_kv_store(self, key: str):
        self.temp_kv_store.pop(key)

    async def get_kv_store(self, key: str) -> str:
        return self.temp_kv_store.get(key)

    """
    Rollout functionality
    """

    async def get_batched_prompt(
        self,
        n: int,
        validation_step: Optional[int] = None,
    ) -> Tuple[List[IdxAndRLPayload], bool]:
        add_answer = (
            self.config.rollout.multi_turn_config.enable
            or not self.config.rollout.reference_answer_in_local
        )

        # query n prompts from the dataset [idx, payload]
        prompt_id_and_payload_list: List[IdxAndRLPayload] = []
        is_end = False

        is_validation = validation_step is not None
        if is_validation:
            iterator = self.policy_status_manager.validation_get_dataloader(
                validation_step
            )
            batch_size = self.val_batch_size
        else:
            iterator = self.train_dataloader_iter
            batch_size = self.rollout_batch_size

        if not is_validation:
            # Throttle the generation speed:
            # 1. Detect the current left pending rollouts in all policy replicas.
            # 2. Check the config.train.train_policy.allowed_outdated_steps.
            # 3. If the current pending rollouts is larger than the allowed outdated version count, reduce the number of prompts to generate.
            current_pending_rollouts = (
                self.policy_status_manager.total_pending_rollouts()
            )
            if (
                current_pending_rollouts
                > self.config.train.train_policy.allowed_outdated_steps
                * len(self.policy_status_manager)
                * self.config.train.train_batch_per_replica
            ):
                n = 0 if self.config.train.train_policy.no_outdated_rollout else 1
                if not self.config.train.train_policy.no_outdated_rollout:
                    # Log only when n is reduced but not when set to 0 since 0 is logged too frequently
                    logger.warning(
                        f"[Controller] Current pending rollouts {current_pending_rollouts} is larger than the allowed outdated version count {self.config.train.train_policy.allowed_outdated_steps * len(self.policy_status_manager)}. Generate with batch {n}"
                    )

        def _next_payload(
            iterator, add_answer: bool
        ) -> tuple[List[int], List[RLPayload]]:
            idxs, payloads = next(iterator)
            assert len(idxs) <= batch_size
            assert len(payloads) <= batch_size
            assert len(idxs) == len(payloads)
            updated_payloads: List[RLPayload] = []
            for idx, payload in zip(idxs, payloads):
                if add_answer:
                    if is_validation:
                        payload.reference_answer = (
                            self.val_dataset.val_set.get_reference_answer(idx)
                        )
                    else:
                        payload.reference_answer = (
                            self.dataset.train_set.get_reference_answer(idx)
                        )
                updated_payloads.append(payload)
            return idxs, updated_payloads

        for _ in range(math.ceil(n / batch_size)):
            payload: RLPayload | None = None
            try:
                idxs, payloads = _next_payload(iterator, add_answer)
            except StopIteration:
                if not is_validation:
                    self.epoch += 1
                    if self.epoch <= self.config.train.epoch:
                        logger.info(f"[Controller] Epoch {self.epoch} start.")
                        iterator = iter(self.train_dataloader)
                        self.train_dataloader_iter = iterator

                        idxs, payloads = _next_payload(iterator, add_answer)
                    else:
                        if self.epoch == self.config.train.epoch + 1:
                            # We only log this all finished information once.
                            logger.info(
                                "[Controller] All epochs finished fetching rollout prompts, wait for rollouts generation and training to complete."
                            )
                        is_end = True
                        break
                else:
                    is_end = True
                    break
            assert len(idxs) == len(payloads)
            for idx, payload in zip(idxs, payloads):
                idx = idx.item() if isinstance(idx, torch.Tensor) else idx
                prompt_id_and_payload_list.append((idx, payload))

        current_fetch_count = len(prompt_id_and_payload_list)
        if (
            (not is_validation)
            and self.config.train.train_policy.on_policy
            and len(self.rollout_status_manager.replica_scaling_log) == 0
        ):
            # Fully Synchronized mode is enabled, we need to tag the prompt with specific weight-version
            global_batch_size = (
                self.config.train.train_batch_per_replica
                * len(self.policy_status_manager)
                // self.config.rollout.n_generation
            )  # global_batch_size: number of prompts needed for single policy step.
            num_of_valid_prompts_consumed = (
                self.policy_status_manager.consumed_samples_num
                // self.config.rollout.n_generation
            )
            weight_version_for_current_batch = (
                num_of_valid_prompts_consumed // global_batch_size
            )

            # record the number of valid prompts for current weight version
            if (
                weight_version_for_current_batch
                not in self.weight_version_to_prompt_num
            ):
                self.weight_version_to_prompt_num[weight_version_for_current_batch] = (
                    current_fetch_count
                )
            else:
                self.weight_version_to_prompt_num[weight_version_for_current_batch] += (
                    current_fetch_count
                )

            # check if for current weight version, we have reached the upper limit of retries to generate enough samples.
            if self.config.train.train_policy.max_retry_for_on_policy > 0:
                already_retried_times = math.ceil(
                    self.weight_version_to_prompt_num[weight_version_for_current_batch]
                    / global_batch_size
                )
                if (
                    already_retried_times
                    > self.config.train.train_policy.max_retry_for_on_policy
                ):
                    raise RuntimeError(
                        f"[Controller] After {self.config.train.train_policy.max_retry_for_on_policy} retries, samples for weight version {weight_version_for_current_batch} are still not enough. May be the dataset is too difficult for current model? Or you could also set the `max_retry_for_on_policy` to 0 or negative to always retry."
                    )

            for i in range(current_fetch_count):
                # get_batched_prompt is called in single thread, so we use `consumed_samples_num` to calculate the weight version.
                # This could ensure that each step of policy will get enough prompts to generae rollouts needed.
                prompt_id_and_payload_list[i][
                    1
                ].weight_version = weight_version_for_current_batch
            # logger.info(f"[Controller] Fully Synchronized mode is enabled, weight_versions: {weight_versions}, train_batch_per_replica: {self.config.train.train_batch_per_replica}, policy_replicas: {len(self.policy_status_manager)}")
        else:
            for i in range(current_fetch_count):
                prompt_id_and_payload_list[i][1].weight_version = 0

        return prompt_id_and_payload_list, is_end

    def query_reference_answer(
        self, prompt_idx: int, dataset_type: str = "train"
    ) -> Any:
        if dataset_type == "train":
            return self.dataset.train_set.get_reference_answer(prompt_idx)
        elif dataset_type == "val":
            return self.val_dataset.val_set.get_reference_answer(prompt_idx)
        else:
            raise ValueError(f"Unknown dataset type: {dataset_type}")

    async def set_profile(self, request: SetProfileRequest):
        replica = self.policy_status_manager[request.replica_name]
        if replica is None:
            logger.warning(
                f"[Controller] Replica {request.replica_name} not found in policy replicas. The profile request takes no effect."
            )
            return {
                "message": "Replica not found in policy replicas. The profile request takes no effect."
            }
        if replica.sub_profiler_config.do_profile:
            logger.warning(
                f"[Controller] Replica {request.replica_name} is already in profile mode. The profile request takes no effect."
            )
            return {
                "message": "Replica is already in profile mode. The profile request takes no effect."
            }
        else:
            kwargs_dict = request.model_dump()
            # remove the replica_name from the kwargs_dict
            kwargs_dict.pop("replica_name")
            # add do_profile to the kwargs_dict
            kwargs_dict["do_profile"] = True
            replica.sub_profiler_config = SubProfilerConfig(**kwargs_dict)
            logger.info(
                f"[Controller] Set profile mode for replica {request.replica_name}."
            )
            return {"message": f"Set replica {request.replica_name} to profile mode."}

    async def set_trace_path(
        self, replica_name: str, trace_path: str, global_rank: int
    ):
        replica = self.policy_status_manager[replica_name]
        if replica is None:
            logger.warning(
                f"[Controller] Replica {replica_name} not found in policy replicas. The trace path request takes no effect."
            )
            return None
        return await replica.set_trace_path(trace_path, global_rank)

    async def put_rollouts(
        self, valid_rollouts: List[Rollout], invalid_rollouts: List[Rollout]
    ):
        """
        Dispatch the rollouts to the policy replicas in a round-robin manner.
        valid_rollouts: List[Rollout]: The rollouts that have valid rewards
        invalid_rollouts: List[Rollout]: The rollouts that have invalid rewards (all rewards are the same)
        """
        completion_tokens_count, n_samples = self.policy_status_manager.put_rollouts(
            valid_rollouts, invalid_rollouts
        )

        self.stat_completion_tokens_count += completion_tokens_count
        self.stat_n_samples += n_samples

        # Statistic
        if self.begin_time is None:
            self.begin_time = time.time()

        # Print pending rollouts inside all policy replicas
        pending_count = self.policy_status_manager.total_pending_rollouts()

        elapsed_time_in_seconds = time.time() - self.begin_time
        logger.info(
            f"[Controller] Stat: {self.stat_n_samples} samples, {self.stat_completion_tokens_count} completion tokens, {pending_count} pending rollouts, {elapsed_time_in_seconds} seconds elapsed"
        )

    """
    State of controller
    """

    def policy_mesh_and_group_size(self) -> tuple[List[str], List[int]]:
        mesh_names = copy.deepcopy(MESH_NAMES)
        group_sizes = []
        for replica in self.policy_status_manager:
            group_sizes.append(replica.group_size)
            break

        return mesh_names, group_sizes

    def rollout_mesh_and_group_size(self) -> tuple[List[str], List[int]]:
        mesh_names = copy.deepcopy(MESH_NAMES)
        group_sizes = []
        for replica in self.rollout_status_manager:
            group_sizes.append(replica.group_size)
            break

        return mesh_names, group_sizes

    def replica_heartbeat(self, replica_name: str):
        if replica_name in self.policy_status_manager:
            self.policy_status_manager.heartbeat(replica_name)
        elif replica_name in self.rollout_status_manager:
            self.rollout_status_manager.heartbeat(replica_name)
        else:
            logger.error(f"[Controller] Replica {replica_name} not found")

    """
    Life-cycle of controller
    """

    async def register(self, atom: Atom, role: Role):
        async with self.life_cycle_lock:
            if role == Role.POLICY:
                self.policy_status_manager.register(
                    atom, self.config, self.rollout_status_manager
                )
            elif role == Role.ROLLOUT:
                self.rollout_status_manager.register(
                    atom, self.config, self.policy_status_manager
                )
            else:
                raise Exception(f"[Controller] Unknown role: {role}")

    async def unregister(self, replica_name: str):
        logger.info(f"[Controller] Unregistering replica {replica_name}")
        async with self.life_cycle_lock:
            if replica_name in self.policy_status_manager:
                self.policy_status_manager.unregister(replica_name)
            elif replica_name in self.rollout_status_manager:
                self.rollout_status_manager.unregister(
                    replica_name, self.policy_status_manager
                )
            else:
                raise Exception(f"[Controller] Replica {replica_name} not found")

    async def set_replica_ncclerror(self, replica_name: str, error: str):
        if replica_name in self.policy_status_manager:
            self.policy_status_manager.set_ncclerror(replica_name, int(time.time()))

            # we use a time window to check nccl report, the last report will invoke post_ncclerror
            self.post_ncclerror_policy_invoke_id += 1
            current_invoke_id = self.post_ncclerror_policy_invoke_id
            await asyncio.sleep(constant.COSMOS_NCCL_ERROR_CLEAN_REPLICA_DELAY)
            if current_invoke_id == self.post_ncclerror_policy_invoke_id:
                # only the latest invoke will trigger the nccl error check
                await self.post_ncclerror(
                    self.policy_status_manager.get_all_policy_report_ncclerror(),
                    Role.POLICY,
                )
                self.policy_status_manager.clear_ncclerror()
        elif replica_name in self.rollout_status_manager:
            raise NotImplementedError(
                f"[Controller] Rollout replica {replica_name} set timeout ack not supported"
            )
        else:
            logger.error(
                f"[Controller] Replica {replica_name} not found in both policy and rollout."
            )

    async def post_ncclerror(
        self, replicas_report_ncclerror: Dict[str, int], role: Role
    ):
        """
        This function is used to clean the hang replicas and trigger the buildmesh command
        """
        all_replicas_ = (
            self.policy_status_manager.policy_replicas
            if role == Role.POLICY
            else self.rollout_status_manager.rollout_replicas
        )
        live_replicas = {rn: all_replicas_[rn] for rn in replicas_report_ncclerror}
        hang_replicas = [
            replica_name
            for replica_name in all_replicas_
            if replica_name not in live_replicas
        ]

        logger.info(f"[Controller] will clean hang replicas: {hang_replicas}")

        if len(live_replicas) == 1:
            # if there is only one replica, it's critical status, we should warning user to scale up the replica
            logger.warning(
                "[Controller] Only one replica is live, it's critical status, user should scale up the replica ASAP!"
            )

        # step 1, manual unregister the hang replicas, we only trigger buildmesh command after update the status
        if role == Role.POLICY:
            for hang_replica in hang_replicas:
                self.policy_status_manager.unregister(hang_replica)
        elif role == Role.ROLLOUT:
            raise NotImplementedError(
                f"[Controller] Rollout replica {hang_replica} set timeout ack not supported"
            )
        else:
            raise Exception(f"[Controller] Unknown role during post_ncclerror: {role}")



from typing import Dict, Tuple

import torch
from torch.nn import Parameter

from vllm.model_executor.layers.quantization.mxfp4 import Mxfp4MoEMethod
from vllm.model_executor.layers.fused_moe import FusedMoE

from cosmos_rl.utils.arithmetic import cdiv
from cosmos_rl.policy.model import WeightMapper
from cosmos_rl.utils.parallelism import ParallelDims

"""
This file is used to patch the vllm model to use mxfp4 for GPT-OSS now.
"""


def quantize_mx4(w):
    from triton_kernels.numerics_details.mxfp import downcast_to_mxfp

    w, w_scale = downcast_to_mxfp(w.to(torch.bfloat16), torch.uint8, axis=1)
    # w = convert_layout(wrap_torch_tensor(w, dtype=FP4), HopperMXValueLayout, mx_axis=1)
    # w_scale = convert_layout(wrap_torch_tensor(w_scale), StridedLayout)
    return w, w_scale


def replace_weight_of_quantized_module(
    vllm_model: torch.nn.Module,
    cached_weight_map: Dict[str, torch.Tensor],
    weight_mapper: WeightMapper,
):
    """
    Temporarily replace the quantized fp8 layer's weight with the cached weight.
    """
    for name, module in vllm_model.named_modules():
        # Here we use the compatible name as the key, aligned with what we do in
        # `cache_weight_of_quantized_module` and `rollout_prepare_recv`.
        # create a new key for gate_up_proj and down_proj
        w13_weight_name = "w13_weight"
        w2_weight_name = "w2_weight"
        w13_bias_name = "w13_bias"
        for weight_name in [w13_weight_name, w2_weight_name, w13_bias_name]:
            compatible_name = weight_mapper._rollout_vllm_name_to_hf(
                name + f".{weight_name}"
            )
            if compatible_name in cached_weight_map:
                setattr(module, weight_name, cached_weight_map[compatible_name])


def cache_weight_of_quantized_module(
    vllm_model: torch.nn.Module,
    promotion_dtype: torch.dtype,
    weight_mapper: WeightMapper,
    parallel_dims: ParallelDims,
) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
    """Get the weight from the quantized module."""
    original_weight_map = {}
    hp_weight_map = {}

    for name, module in vllm_model.named_modules():
        quant_method = getattr(module, "quant_method", None)
        if quant_method is None:
            continue
        elif isinstance(quant_method, Mxfp4MoEMethod):
            assert isinstance(
                module, FusedMoE
            ), "Mxfp4MoEMethod should be used with FusedMoE"
            # w13 and w2
            w13_weight_name = "w13_weight"
            w13_compatible_name = weight_mapper._rollout_vllm_name_to_hf(
                name + f".{w13_weight_name}"
            )
            w2_weight_name = "w2_weight"
            w2_compatible_name = weight_mapper._rollout_vllm_name_to_hf(
                name + f".{w2_weight_name}"
            )
            w13_bias_name = "w13_bias"
            w13_bias_compatible_name = weight_mapper._rollout_vllm_name_to_hf(
                name + f".{w13_bias_name}"
            )
            # Keep same as vLLM did: delete the original weight.
            original_weight_map[w13_compatible_name] = None
            original_weight_map[w2_compatible_name] = None
            original_weight_map[w13_bias_compatible_name] = module.w13_bias

            # We assume that MoE high precision weight should have shape:
            # w13:[num_experts, 2 * intermediate_size, hidden_size]
            # w2: [num_experts, hidden_size, intermediate_size]
            # FIXME: (lms) So we just don't support EP for temporarily. When vLLM support EP for gpt-oss, fix this.
            # intermediate_size_per_partition = (
            #     module.intermediate_size_per_partition
            # )  # equals to: intermediate_size // self.tp_size
            hidden_size = module.hidden_size
            w13_bias = getattr(module, "w13_bias", None)
            assert w13_bias is not None, "w13_bias should exist"
            # Note: we assume that Policy HFModel is in BF16.

            intermediate_size = weight_mapper.config.intermediate_size
            mxfp4_block = 32
            intermediate_size_block = intermediate_size // mxfp4_block  # 90
            tp_rank, tp_size = parallel_dims.tp_coord
            per_rank_intermediate_size_block = cdiv(intermediate_size_block, tp_size)
            per_rank_intermediate_size = per_rank_intermediate_size_block * mxfp4_block

            tp_rank_start = tp_rank * per_rank_intermediate_size
            tp_rank_end = min(
                (tp_rank + 1) * per_rank_intermediate_size, intermediate_size
            )

            length = tp_rank_end - tp_rank_start

            w13_hp_weight = torch.empty(
                [
                    module.local_num_experts,
                    2 * length,
                    hidden_size,
                ],
                dtype=promotion_dtype,
                device=w13_bias.device,
            )
            w2_hp_weight = torch.empty(
                [
                    module.local_num_experts,
                    hidden_size,
                    length,
                ],
                dtype=promotion_dtype,
                device=w13_bias.device,
            )

            hp_weight_map[w13_compatible_name] = Parameter(
                w13_hp_weight, requires_grad=False
            )
            hp_weight_map[w2_compatible_name] = Parameter(
                w2_hp_weight, requires_grad=False
            )
            # For w13_bias
            w13_bias_hp_weight = torch.empty(
                [
                    module.local_num_experts,
                    2 * length,
                ],
                dtype=w13_bias.dtype,
                device=w13_bias.device,
            )
            hp_weight_map[w13_bias_compatible_name] = Parameter(
                w13_bias_hp_weight, requires_grad=False
            )
        else:
            # We will not handle other quant methods.
            pass
    return hp_weight_map, original_weight_map


def post_process_view_map_for_mxfp4(
    vllm_weight_inplace_view_map: Dict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    """Process the view map returned by `rollout_prepare_recv`.
            - remove the weight_scale from the view map.
    Args:
        vllm_weight_inplace_view_map (Dict[str, torch.Tensor]): view map returned by `rollout_prepare_recv`
    Returns:
        Dict[str, torch.Tensor]: view map doesn't contain weight_scale.
    """
    processed_view_map = {}
    for key, value in vllm_weight_inplace_view_map.items():
        if "down_proj_scales" in key or "gate_up_proj_scales" in key:
            continue
        processed_view_map[key] = value
    return processed_view_map



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.policy.trainer import Trainer
from cosmos_rl.utils.parallelism import (
    ParallelDims,
)
from cosmos_rl.policy.config import (
    Config as CosmosConfig,
    SFTDataConfig,
    config_hash,
)
from cosmos_rl.policy.trainer.optm import build_lr_schedulers
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.wandb_logger import (
    init_wandb,
    is_wandb_available,
    log_wandb,
)
import torch
import numpy as np
from torch.utils.data import Dataset, DataLoader, DistributedSampler, Sampler
from cosmos_rl.policy.trainer.sampler import SkippingSampler
import cosmos_rl.utils.util as util
import cosmos_rl.utils.distributed as dist_util
import cosmos_rl.utils.cache as cache
from transformers import AutoTokenizer
from datasets import concatenate_datasets
from cosmos_rl.dispatcher.data.packer import DataPacker
import os
from typing import Optional, Dict, Any, Callable, Union
from tqdm import tqdm
from cosmos_rl.utils.ulysses import (
    slice_inputs_for_ulysses,
)
from functools import partial
from cosmos_rl.utils.sequence_packing import (
    pack_sequences_info_collect,
    pack_sequences_for_masks,
    pack_sequences_for_labels,
)
from itertools import islice


def async_safe_ce(
    output: torch.Tensor,
    target: torch.LongTensor,
    ignore_index: int = -100,
    loss_scaling_factor: float = 1.0,
    output_packing_mask: Optional[torch.Tensor] = None,
    target_packing_mask: Optional[torch.Tensor] = None,
    dp_group: Optional[torch.distributed.ProcessGroup] = None,
    cp_group: Optional[torch.distributed.ProcessGroup] = None,
) -> torch.Tensor:
    if output_packing_mask is not None:
        output = (
            output[output_packing_mask].contiguous().view(-1, output.size(-1)).float()
        )
    else:
        output = output[:, :-1].contiguous().view(-1, output.size(-1)).float()

    if target_packing_mask is not None:
        target = target[target_packing_mask].contiguous().view(-1)
    else:
        target = target[:, 1:].contiguous().view(-1)

    if cp_group is not None and cp_group.size() > 1:
        # Fallback to unbalance loss
        loss = (
            torch.nn.functional.cross_entropy(
                output,
                target,
                ignore_index=ignore_index,
                reduction="mean",
            )
            * loss_scaling_factor
        )
        # In case of all labels are ignored, loss will be nan.
        loss = torch.nan_to_num(loss, nan=0.0)
        return loss
    else:
        loss = torch.nn.functional.cross_entropy(
            output,
            target,
            ignore_index=ignore_index,
            reduction="none",
        )

        # Compute all token numbers across dp-world
        n_valid_tokens = (target != ignore_index).sum()
        num_dp_workers = 1
        if dp_group is not None:
            torch.distributed.all_reduce(n_valid_tokens, group=dp_group)
            num_dp_workers = torch.distributed.get_world_size(group=dp_group)

        loss = (
            loss.sum()
            / (n_valid_tokens + 1e-8)
            * (num_dp_workers * loss_scaling_factor)
        )
        return loss


def collate_fn(
    batch,
):
    return batch


def construct_dataset(
    cosmos_config: CosmosConfig,
    tokenizer: AutoTokenizer,
    data_packer: DataPacker,
    user_provided_dataset: Optional[Dataset] = None,
    val_data_packer: Optional[DataPacker] = None,
    user_provided_val_dataset: Optional[Dataset] = None,
):
    config = cosmos_config.train.train_policy
    if user_provided_dataset is not None:
        dataset = None
        train_dataset = user_provided_dataset
        logger.info("Using user-provided dataset, which will skip split processing.")
    else:
        dataset = util.load_data_from_disk_or_hf(
            config.dataset.name,
            config.dataset.subset,
            config.dataset.revision or None,
        )
        dataset_list = []
        for split_name in config.dataset.split:
            logger.info(
                f"Appending split {split_name}, dataset size = {len(dataset[split_name])}"
            )
            dataset_list.append(dataset[split_name])
        train_dataset = concatenate_datasets(dataset_list)
    logger.info(f"Final dataset size = {len(train_dataset)}")

    if cosmos_config.validation.enable:
        if user_provided_val_dataset is not None:
            test_dataset = user_provided_val_dataset
            logger.info(
                "Using user-provided validation dataset, which will skip split processing."
            )
        elif cosmos_config.validation.dataset.name:
            dataset = util.load_data_from_disk_or_hf(
                cosmos_config.validation.dataset.name,
                cosmos_config.validation.dataset.subset,
                cosmos_config.validation.dataset.revision or None,
            )
            dataset_list = []
            for split_name in cosmos_config.validation.dataset.split:
                logger.info(
                    f"Appending validation split {split_name}, validation dataset size = {len(dataset[split_name])}"
                )
                dataset_list.append(dataset[split_name])
            test_dataset = concatenate_datasets(dataset_list)
        else:
            logger.warning(
                "No validation dataset provided, using split of training dataset for validation."
            )
            if isinstance(train_dataset, torch.utils.data.Dataset):
                # Define the split ratio (e.g., 80% train, 20% test)
                if config.dataset.test_size is None:
                    logger.warning(
                        "No test size specified, using 10% of the training dataset for testing."
                    )
                    config.dataset.test_size = 0.1
                if isinstance(config.dataset.test_size, float):
                    n_test_samples = int(len(train_dataset) * config.dataset.test_size)
                else:
                    n_test_samples = config.dataset.test_size
                n_test_samples = max(min(n_test_samples, len(train_dataset) - 1), 1)

                # Generate deterministic indices
                indices = list(range(len(train_dataset)))
                test_indices = indices[:n_test_samples]
                train_indices = indices[n_test_samples:]

                test_dataset = torch.utils.data.Subset(train_dataset, test_indices)
                train_dataset = torch.utils.data.Subset(train_dataset, train_indices)
            else:
                assert hasattr(
                    train_dataset, "train_test_split"
                ), "train_dataset must have train_test_split method"
                split = train_dataset.train_test_split(
                    test_size=config.dataset.test_size, shuffle=False
                )
                train_dataset = split["train"]
                test_dataset = split["test"]
    else:

        class EmptyDataset(Dataset):
            def __len__(self):
                return 0

            def __getitem__(self, idx):
                raise IndexError("EmptyDataset has no items")

        test_dataset = EmptyDataset()

    train_sft_dataset = SFTDataset(
        config,
        tokenizer=tokenizer,
        dataset=train_dataset,
        data_packer=data_packer,
        is_user_dataset=user_provided_dataset is not None,
    )
    test_sft_dataset = SFTDataset(
        config,
        tokenizer=tokenizer,
        dataset=test_dataset,
        data_packer=val_data_packer,
        is_user_dataset=user_provided_dataset is not None,
    )

    return train_sft_dataset, test_sft_dataset


class SFTDataset(Dataset):
    def __init__(
        self,
        config: SFTDataConfig,
        tokenizer: AutoTokenizer,
        dataset: Dataset,
        data_packer: DataPacker,
        is_user_dataset: bool = False,
    ):
        self.config = config
        self.tokenizer = tokenizer
        self.column_name = config.conversation_column_name
        self.dataset = dataset
        self.data_packer = data_packer
        self.is_user_dataset = is_user_dataset
        self.cache = None
        if self.config.enable_dataset_cache:
            # TODO(zjx): can we reuse the cache between different training jobs?
            # It's not stable yet, we only checked if the config is the same
            # If there are any problems, it is recommended that the user clears the cache folder
            cache_folder = os.path.join(
                os.environ.get(
                    "COSMOS_CACHE",
                    os.path.join(os.path.expanduser("~"), ".cache/cosmos/"),
                ),
                "datasets_cache",
                f"{self.config.dataset.name}-{config_hash(config)}",
            )
            logger.info(f"SFTDataset Cache folder: {cache_folder}")
            self.cache = cache.DiskCache(cache_folder)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        # we only cache on_the_fly result
        if self.cache is not None:
            cache_obj = self.cache.get(idx)
            if cache_obj is not None:
                return cache_obj

        raw_item = (
            self.dataset[idx][self.column_name]
            if not self.is_user_dataset and self.column_name
            else self.dataset[idx]
        )

        if isinstance(idx, list):  # a batch of items
            item = [self.data_packer.sft_process_sample(x) for x in raw_item]
        else:
            item: Dict[str, Any] = self.data_packer.sft_process_sample(raw_item)

        if self.cache is not None:
            # try cache obj
            self.cache.set(idx, item)
        return item


class SFTTrainer(Trainer):
    def __init__(
        self,
        config: CosmosConfig,
        parallel_dims: ParallelDims,
        dataset: Optional[Dataset] = None,
        data_packer: Optional[DataPacker] = None,
        val_dataset: Optional[Dataset] = None,
        val_data_packer: Optional[DataPacker] = None,
        sampler: Optional[Callable] = None,
        batch_sampler: Optional[Callable] = None,
        val_sampler: Optional[Callable] = None,
        val_batch_sampler: Optional[Callable] = None,
    ):
        super(SFTTrainer, self).__init__(config, parallel_dims)

        # Enlarge the compile cache size for validation
        if config.train.compile and config.validation.enable:
            torch._dynamo.config.cache_size_limit = 64

        self.dp_rank, self.dp_world_size = 0, 1
        if parallel_dims.dp_enabled:
            self.dp_rank = parallel_dims.mesh["dp"].get_local_rank()
            self.dp_world_size = parallel_dims.mesh["dp"].size()

        # Prepare wandb
        if "wandb" in config.logging.logger and is_wandb_available():
            init_wandb(config, parallel_dims)
        else:
            logger.warning(
                "Wandb is not available. Please install it to use wandb logging features."
            )

        self.train_step = 0
        ckpt_total_steps = 0
        self.lr_schedulers = None
        self.start_epoch = 0
        # Load model
        if config.train.resume:
            try:
                # early init the lr_schedulers to avoid it is not initialized when loading the checkpoint
                ckpt_extra_vars, self.lr_schedulers = self.ckpt_manager.load_checkpoint(
                    model=self.model,
                    optimizer=self.optimizers,
                    scheduler=partial(
                        build_lr_schedulers, self.optimizers, self.config
                    ),
                )
                ckpt_total_steps = ckpt_extra_vars.get("total_steps", 0)
                self.train_step = ckpt_extra_vars.get("step", 0)
            except Exception as e:
                logger.error(
                    f"Cannot resume due to error: {e}. Trying to load from HuggingFace..."
                )
                self.model.load_hf_weights(
                    config.policy.model_name_or_path,
                    parallel_dims,
                    self.device,
                    revision=config.policy.model_revision,
                )
        else:
            self.model.load_hf_weights(
                config.policy.model_name_or_path,
                parallel_dims,
                self.device,
                revision=config.policy.model_revision,
            )
        self.model.train()

        if isinstance(dataset, Callable):
            # Incase it is a factory function, we need to call it to get the dataset
            dataset = dataset(self.config)
            dataset.setup(self.config, self.tokenizer)
        if data_packer:
            data_packer.setup(self.config, self.tokenizer)
            self.data_packer = data_packer

        if isinstance(val_dataset, Callable):
            val_dataset = val_dataset(self.config)
            val_dataset.setup(self.config, self.tokenizer)
        if val_data_packer:
            val_data_packer.setup(self.config, self.tokenizer)
            self.val_data_packer = val_data_packer
        else:
            self.val_data_packer = self.data_packer

        # Prepare dataset
        train_dataset, val_dataset = construct_dataset(
            config,
            tokenizer=self.tokenizer,
            data_packer=self.data_packer,
            user_provided_dataset=dataset,
            val_data_packer=self.val_data_packer,
            user_provided_val_dataset=val_dataset,
        )
        if sampler is not None:
            logger.info("Using user-provided sampler for training dataset.")
            if isinstance(sampler, Callable):
                train_sampler = sampler(
                    train_dataset,
                    num_replicas=self.dp_world_size,
                    rank=self.dp_rank,
                    shuffle=config.train.train_policy.dataloader_shuffle,
                    drop_last=False,
                )
            else:
                train_sampler = sampler
        else:
            train_sampler = DistributedSampler(
                train_dataset,
                num_replicas=self.dp_world_size,
                rank=self.dp_rank,
                shuffle=config.train.train_policy.dataloader_shuffle,
                drop_last=False,
            )

        if batch_sampler is not None and isinstance(batch_sampler, Callable):
            batch_sampler = batch_sampler(
                train_sampler,
                batch_size=config.train.train_batch_per_replica,
                drop_last=False,
            )

        def get_train_data_loader(
            sampler: Union[Sampler[int], Sampler[list[int]]],
            sampler_in_batch: Optional[Sampler[list[int]]] = None,
        ):
            if sampler_in_batch is not None:
                logger.info(
                    "Using custom batch Sampler that yields list of indices for training dataset."
                )
                data_loader = DataLoader(
                    train_dataset,
                    num_workers=config.train.train_policy.dataloader_num_workers,
                    prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                    batch_sampler=sampler_in_batch,
                    collate_fn=collate_fn,
                )
            else:
                data_loader = DataLoader(
                    train_dataset,
                    batch_size=config.train.train_batch_per_replica,
                    shuffle=False,
                    num_workers=config.train.train_policy.dataloader_num_workers,
                    prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                    sampler=sampler,
                    collate_fn=collate_fn,
                    drop_last=False,
                )
            return data_loader

        if config.train.resume and self.train_step > 0:
            """
            Note: Here we assume there is no data shuffling across epochs.
            Otherwise, we need to call `set_epoch` on the sampler after each epoch.
            """
            # Resume training from the last checkpoint if needed
            total_steps_per_epoch = len(
                get_train_data_loader(train_sampler, batch_sampler)
            )
            data_loader_bias = self.train_step % total_steps_per_epoch
            data_loader_bias *= config.train.train_batch_per_replica
            logger.info(
                f"Resuming training from step {self.train_step}/{ckpt_total_steps}"
            )
            train_sampler = SkippingSampler(
                train_sampler,
                skip_samples=data_loader_bias
                // (
                    len(list(islice(iter(train_sampler), 1))[0])
                    if isinstance(list(islice(iter(train_sampler), 1))[0], list)
                    else 1
                ),
            )
            if batch_sampler is not None:
                batch_sampler = SkippingSampler(
                    batch_sampler,
                    skip_samples=data_loader_bias
                    // (
                        len(list(islice(iter(batch_sampler), 1))[0])
                        if isinstance(list(islice(iter(batch_sampler), 1))[0], list)
                        else 1
                    ),
                )
            self.start_epoch = self.train_step // total_steps_per_epoch

        if val_sampler is not None:
            logger.info("Using user-provided sampler for validation dataset.")
            if isinstance(val_sampler, Callable):
                val_sampler = val_sampler(
                    val_dataset,
                    num_replicas=self.dp_world_size,
                    rank=self.dp_rank,
                    shuffle=False,
                    drop_last=False,
                )
        else:
            val_sampler = DistributedSampler(
                val_dataset,
                num_replicas=self.dp_world_size,
                rank=self.dp_rank,
                shuffle=False,
                drop_last=False,
            )
        self.epoch = config.train.epoch

        assert (
            self.tokenizer.pad_token_id is not None
        ), "Tokenizer must have a pad token id"
        self.train_data_loader = get_train_data_loader(train_sampler, batch_sampler)
        if val_batch_sampler is not None:
            logger.info(
                "Using custom batch Sampler that yields list of indices for validation dataset."
            )
            if isinstance(val_batch_sampler, Callable):
                val_batch_sampler = val_batch_sampler(
                    val_sampler,
                    batch_size=config.validation.batch_size
                    or config.train.train_batch_per_replica,
                    drop_last=False,
                )
            self.val_data_loader = DataLoader(
                val_dataset,
                num_workers=config.train.train_policy.dataloader_num_workers,
                prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                batch_sampler=val_batch_sampler,
                collate_fn=collate_fn,
            )
        else:
            self.val_data_loader = DataLoader(
                val_dataset,
                batch_size=config.validation.batch_size
                or config.train.train_batch_per_replica,
                num_workers=config.train.train_policy.dataloader_num_workers,
                prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                sampler=val_sampler,
                collate_fn=collate_fn,
                drop_last=False,
            )

        steps_by_dataset = (
            ckpt_total_steps
            if ckpt_total_steps > 0
            else len(self.train_data_loader) * self.epoch
        )
        if config.train.max_num_steps is not None:
            self.total_steps = min(steps_by_dataset, config.train.max_num_steps)
        else:
            self.total_steps = steps_by_dataset

        if self.lr_schedulers is None:
            assert (
                self.train_step == 0
            ), "`SFTTrainer.lr_schedulers` should be None if training is from scratch"
            self.lr_schedulers = build_lr_schedulers(
                self.optimizers, self.config, self.total_steps
            )

        if self.parallel_dims.dp_shard_enabled:
            dp_group = self.parallel_dims.mesh["dp_shard"].get_group()
        else:
            dp_group = None

        if self.parallel_dims.cp_enabled:
            cp_group = self.parallel_dims.mesh["cp"].get_group()
        else:
            cp_group = None

        self.loss_fn = partial(
            async_safe_ce,
            dp_group=dp_group
            if self.config.train.train_policy.balance_dp_token
            else None,
            cp_group=cp_group,
        )

        # Calculate the step interval to save the checkpoint
        if self.config.train.ckpt.save_freq_in_epoch > 0:
            # Use save_freq_in_epoch to calculate the save frequency in priority
            self._save_freq = (
                self.config.train.ckpt.save_freq_in_epoch * len(self.train_data_loader)
            ) // self.dp_world_size
            logger.info(
                f"Checkpoint will be saved every {self._save_freq} steps, which is approximately every `train.ckpt.save_freq_in_epoch` {self.config.train.ckpt.save_freq_in_epoch} epochs. `train.ckpt.save_freq` will be ignored."
            )
        else:
            self._save_freq = self.config.train.ckpt.save_freq

    def validate(self):
        if not self.config.validation.enable:
            return
        if self.parallel_dims.dp_replicate_coord[0] != 0:
            return

        logger.info(f"Validation at step {self.train_step}/{self.total_steps}...")
        self.model.eval()
        with torch.no_grad():
            val_total_loss = 0.0
            for val_global_batch in tqdm(self.val_data_loader, desc="Validation"):
                fixed_length = (
                    self.config.policy.model_max_length
                    if self.parallel_dims.pp_enabled
                    and not self.parallel_dims.pp_dynamic_shape
                    else None
                )
                if fixed_length is None:
                    max_len = min(
                        self.config.policy.model_max_length,
                        self.val_data_packer.sft_compute_max_len(val_global_batch),
                    )
                else:
                    max_len = fixed_length
                if self.seq_len_multiple > 1:
                    max_len = (
                        (max_len + self.seq_len_multiple - 1)
                        // self.seq_len_multiple
                        * self.seq_len_multiple
                    )

                val_batch = self.val_data_packer.sft_collate_fn(
                    val_global_batch,
                    computed_max_len=max_len,
                    pad_token_id=self.tokenizer.pad_token_id,
                    ignore_label_id=-100,
                )
                for k, v in val_batch.items():
                    val_batch[k] = (
                        v.to(self.device) if isinstance(v, torch.Tensor) else v
                    )
                val_inputs = val_batch["input_ids"]
                val_labels = val_batch.pop("label_ids")
                val_position_ids, _, val_pos_seq_dim = self.model.get_position_ids(
                    **val_batch
                )

                val_batch["position_ids"] = val_position_ids
                val_padding_mask = val_batch.get("padding_mask", None)

                delay_cp_slice_inputs = getattr(
                    self.model, "delay_cp_slice_inputs", False
                )
                if self.parallel_dims.cp_enabled and not delay_cp_slice_inputs:
                    [val_inputs, val_position_ids, val_padding_mask] = (
                        slice_inputs_for_ulysses(
                            [val_inputs, val_position_ids, val_padding_mask],
                            self.parallel_dims.mesh["cp"],
                            seq_dims=[1, val_pos_seq_dim, 1],
                        )
                    )

                    val_batch["input_ids"] = val_inputs
                    val_batch["position_ids"] = val_position_ids
                    if val_padding_mask is not None:
                        val_batch["padding_mask"] = val_padding_mask

                if self.parallel_dims.pp_enabled:
                    pp_last_stage = (
                        self.parallel_dims.pp_coord[0]
                        == self.parallel_dims.pp_coord[1] - 1
                    )
                    pp_first_stage = self.parallel_dims.pp_coord[0] == 0

                    if pp_first_stage:
                        self.pp_scheduler_val.step(
                            **val_batch,
                            pp_dynamic_shape_enabled=self.parallel_dims.pp_dynamic_shape_enabled,
                            seq_len_multiple=self.seq_len_multiple,
                        )
                    else:
                        pp_out = self.pp_scheduler_val.step(
                            position_ids=val_position_ids,
                            pp_dynamic_shape_enabled=self.parallel_dims.pp_dynamic_shape_enabled,
                            seq_len_multiple=self.seq_len_multiple,
                        )

                    if pp_last_stage:
                        val_loss = self.loss_fn(pp_out, val_labels)
                    else:
                        val_loss = torch.tensor([-1.0], device=self.device)
                else:
                    val_logits = self.model(**val_batch)

                    val_loss = self.loss_fn(val_logits, val_labels)
                val_total_loss += val_loss.item() * val_inputs.size(0)
            val_avg_loss = val_total_loss / len(self.val_data_loader.dataset)
            logger.info(f"Validation loss: {val_avg_loss}")
        return val_avg_loss

    def train(self):
        self.profiler.start()
        pp_last_stage = False

        for cur_epoch in range(self.start_epoch, self.epoch):
            logger.info(f"Training epoch {cur_epoch + 1}/{self.epoch}")
            for global_batch in self.train_data_loader:
                acc_loss = torch.zeros(1, device=self.device)
                self.optimizers.zero_grad()
                global_batch_size = len(global_batch)
                # split global_batch into mini_batches
                mini_batch_begin_idxs = list(
                    range(
                        0,
                        global_batch_size,
                        self.config.train.train_policy.mini_batch,
                    )
                )

                start_event = torch.cuda.Event(enable_timing=True)
                end_event = torch.cuda.Event(enable_timing=True)
                start_event.record()

                for i in mini_batch_begin_idxs:
                    fixed_length = (
                        self.config.policy.model_max_length
                        if self.parallel_dims.pp_enabled
                        and not self.parallel_dims.pp_dynamic_shape
                        else None
                    )
                    raw_batch = global_batch[
                        i : i + self.config.train.train_policy.mini_batch
                    ]
                    if fixed_length is None:
                        max_len = min(
                            self.config.policy.model_max_length,
                            self.data_packer.sft_compute_max_len(raw_batch),
                        )
                    else:
                        max_len = fixed_length

                    if self.seq_len_multiple > 1:
                        max_len = (
                            (max_len + self.seq_len_multiple - 1)
                            // self.seq_len_multiple
                            * self.seq_len_multiple
                        )

                    packing_seq = self.config.train.sequence_packing
                    if packing_seq:
                        if self.parallel_dims.pp_enabled:
                            packing_seq = False
                            logger.debug(
                                "[Policy] Packing sequence is disabled due to incompatible dimensions."
                            )

                    batch = self.data_packer.sft_collate_fn(
                        raw_batch,
                        computed_max_len=max_len,
                        pad_token_id=self.tokenizer.pad_token_id,
                        ignore_label_id=-100,
                    )

                    # if [profiler.enable_nsys] is true, cudaProfilerStart() / cudaProfilerStop() are used to trigger nsys capture
                    # settings from [profiler.sub_profiler_config] are reused
                    if (
                        self.config.profiler.enable_nsys
                        and self.profiler.global_rank in self.profiler.rank_filter
                    ):
                        if (
                            self.train_step
                            == self.profiler.wait_steps + self.profiler.warmup_steps
                        ):
                            torch.cuda.cudart().cudaProfilerStart()
                        elif (
                            self.train_step
                            == self.profiler.wait_steps
                            + self.profiler.warmup_steps
                            + self.profiler.active_steps
                        ):
                            torch.cuda.cudart().cudaProfilerStop()

                    self.model.train()
                    for k, v in batch.items():
                        batch[k] = (
                            v.to(self.device) if isinstance(v, torch.Tensor) else v
                        )

                    labels = batch.pop("label_ids")

                    position_ids, input_ids, pos_seq_dim = self.model.get_position_ids(
                        **batch
                    )

                    batch["position_ids"] = position_ids
                    padding_mask = batch.get("padding_mask", None)

                    if packing_seq:
                        # Prepare for the sequence packing information.
                        packed_args = pack_sequences_info_collect(
                            batch["input_ids"],
                            pad_token_id=self.tokenizer.pad_token_id,
                            label_ids=labels,
                            ignore_label_id=-100,
                            seq_len_multiple=self.seq_len_multiple,
                        )
                        batch.update(packed_args)
                        labels = pack_sequences_for_labels(
                            labels, batch["valid_input_len"]
                        )
                        packed_args = pack_sequences_for_masks(
                            batch["valid_input_len"], batch["valid_input_len"]
                        )
                        batch.update(packed_args)
                    # For VLMs, we need to delay the slice of inputs for CP until after the embedding generation in the model forward.
                    delay_cp_slice_inputs = getattr(
                        self.model, "delay_cp_slice_inputs", False
                    )
                    if (
                        self.parallel_dims.cp_enabled
                        and not packing_seq
                        and not delay_cp_slice_inputs
                    ):
                        [input_ids, position_ids, padding_mask] = (
                            slice_inputs_for_ulysses(
                                [input_ids, position_ids, padding_mask],
                                self.parallel_dims.mesh["cp"],
                                seq_dims=[1, pos_seq_dim, 1],
                            )
                        )

                        batch["input_ids"] = input_ids
                        batch["position_ids"] = position_ids
                        if padding_mask is not None:
                            batch["padding_mask"] = padding_mask

                    if self.parallel_dims.cp_enabled:
                        # Slice for cp after embedding generation and sequence packing in the model forward later.
                        batch["cp_mesh"] = self.parallel_dims.mesh["cp"]

                    if self.parallel_dims.pp_enabled:
                        pp_last_stage = (
                            self.parallel_dims.pp_coord[0]
                            == self.parallel_dims.pp_coord[1] - 1
                        )
                        pp_first_stage = self.parallel_dims.pp_coord[0] == 0

                        # Pipeline Parallel forward / backward inside step() call
                        targets, losses = (
                            (labels, []) if pp_last_stage else (None, None)
                        )
                        if pp_first_stage:
                            self.pp_scheduler.step(
                                **batch,
                                pp_dynamic_shape_enabled=self.parallel_dims.pp_dynamic_shape_enabled,
                                seq_len_multiple=self.seq_len_multiple,
                            )
                        else:
                            # FWD + BWD if it is 1F1B-like scheduler
                            self.pp_scheduler.step(
                                position_ids=batch["position_ids"],
                                target=targets,
                                losses=losses,
                                pp_dynamic_shape_enabled=self.parallel_dims.pp_dynamic_shape_enabled,
                                seq_len_multiple=self.seq_len_multiple,
                            )
                        loss = (
                            torch.mean(torch.stack(losses)).to(self.device)
                            if pp_last_stage
                            else torch.tensor([-1.0], device=self.device)
                        )
                    else:
                        # This code is just for debugging purposes, where we can test whether the model can generate tokens correctly
                        # last_token_ids = []
                        # with torch.no_grad():
                        #     N_NEW_TOKENS = 100
                        #     for _ in range(N_NEW_TOKENS):
                        #         if len(last_token_ids) > 0:
                        #             batch["input_ids"] = torch.cat(
                        #                 [batch["input_ids"], last_token_ids[-1]],
                        #                 dim=-1,
                        #             )
                        #             position_ids, _, _ = (
                        #                 self.model.get_position_ids(**batch)
                        #             )
                        #             batch["position_ids"] = position_ids

                        #         logits = self.model(**batch)
                        #         token_ids = torch.argmax(logits[:, -1:, :], dim=-1)
                        #         last_token_ids.append(token_ids)
                        #     if self.global_rank == 0:
                        #         text = ''
                        #         new_last_token_ids = torch.cat(last_token_ids, dim=-1).squeeze(0)
                        #         logger.info(f'{new_last_token_ids=}')
                        #         text = self.tokenizer.decode(new_last_token_ids)
                        #         logger.info(
                        #             f"generated tokens at sample : {text}"
                        #         )
                        # return
                        #########################################################################################

                        with self.act_offloading_ctx_manager:
                            logits = self.model(**batch)

                        loss = self.loss_fn(
                            logits,
                            labels,
                            output_packing_mask=batch.get("input_packing_mask", None),
                            target_packing_mask=batch.get("label_packing_mask", None),
                            loss_scaling_factor=1.0 / len(mini_batch_begin_idxs),
                        )
                        # # Hint FSDP to do all-reduce on the last backward pass
                        # if hasattr(self.model, "set_is_last_backward"):
                        #     print(f"set_is_last_backward: {i == mini_batch_begin_idxs[-1]}")
                        #     self.model.set_is_last_backward(i == mini_batch_begin_idxs[-1])
                        loss.backward()
                    acc_loss += loss.detach()

                """
                Compute the global grad norm on all parameters and then apply
                gradient clipping using the global grad norm.
                """
                all_params = [
                    p
                    for m in [model for model in self.model_parts if model is not None]
                    for p in m.parameters()
                ]
                grad_norm = dist_util.gradient_norm_clipping(
                    all_params,
                    self.config.train.optm_grad_norm_clip,
                    foreach=True,
                    pp_mesh=self.parallel_dims.mesh["pp"]
                    if self.parallel_dims.pp_enabled
                    else None,
                    return_norm_only=(self.config.train.optm_grad_norm_clip <= 0.0),
                )

                self.optimizers.step()
                self.lr_schedulers.step()

                self.train_step += 1

                # Early stop only when max_num_steps is specified
                if (
                    self.config.train.max_num_steps is not None
                    and self.train_step >= self.total_steps
                ):
                    break

                end_event.record()

                if (
                    self.parallel_dims.dp_replicate_enabled
                    or self.parallel_dims.dp_shard_enabled
                    or self.parallel_dims.cp_enabled
                ):
                    global_avg_loss, global_max_loss = (  # noqa: F841
                        dist_util.dist_mean(acc_loss, self.parallel_dims.mesh["dp_cp"]),
                        dist_util.dist_max(acc_loss, self.parallel_dims.mesh["dp_cp"]),
                    )
                else:
                    global_avg_loss = global_max_loss = acc_loss.item()  # noqa: F841

                if self.config.logging.logger:
                    if util.is_master_rank(self.parallel_dims, self.global_rank):
                        # Calculate last iteration time
                        assert end_event.query()
                        iter_time = (
                            start_event.elapsed_time(end_event) / 1000.0
                        )  # in seconds

                        report_data = {
                            "train/iteration_time": iter_time,
                            "train/loss_avg": global_avg_loss,
                            "train/loss_max": global_max_loss,
                            "train/learning_rate": self.lr_schedulers.get_last_lr()[0],
                            "train/grad_norm": grad_norm
                            if grad_norm is not None
                            else -1,
                        }

                        # FIXME(dinghaoy): only compute MFU of rank 0, if enable tp or pp,
                        # it will be inaccurate. Need a reduce for all the metrics.
                        if self.config.logging.report_mfu:
                            mfu = util.compute_mfu(
                                model=self.model,
                                n_tokens=np.prod(input_ids.shape),
                                iter_time=iter_time,
                                num_gpus=self.world_size,
                                dtype=self.config.train.param_dtype,
                            )
                            for k, v in mfu.items():
                                report_data[f"train/{k}"] = v
                        if (
                            "wandb" in self.config.logging.logger
                            and is_wandb_available()
                        ):
                            log_wandb(
                                data=report_data,
                                step=self.train_step,
                            )
                        if "console" in self.config.logging.logger:
                            logger.info(
                                f"Step: {self.train_step}/{self.total_steps}, Loss: {global_avg_loss:.5f}, Grad norm: {grad_norm:.5f}, Learning rate: {self.lr_schedulers.get_last_lr()[0]:.5e}, Iteration time: {iter_time:.2f}s."
                            )

                # For profiling
                self.profiler.step()

                val_score = None
                # validation
                if self.train_step % self.config.validation.freq == 0:
                    val_score = self.validate()

                # save checkpoint
                if (
                    self.config.train.ckpt.enable_checkpoint
                    and self.train_step % self._save_freq == 0
                    and self.train_step > 0
                    and self.parallel_dims.dp_replicate_coord[0] == 0
                ):
                    # TODO(dinghaoy): support export safetensors asynchronously.
                    if self.config.train.ckpt.export_safetensors:
                        logger.info(
                            f"Saving huggingface checkpoint at step {self.train_step} to {self.config.train.output_dir}..."
                        )
                        self.export_safetensors(
                            output_dir=self.config.train.output_dir,
                            rel_path=os.path.join(
                                "safetensors",
                                f"step_{self.train_step}",
                            ),
                            trainable_only=False,
                            dtype=util.str2torch_dtype(self.config.train.param_dtype),
                        )
                    logger.info(
                        f"Saving cosmos checkpoint at step {self.train_step}..."
                    )
                    self.ckpt_manager.save_checkpoint(
                        model=self.model,
                        optimizer=self.optimizers,
                        scheduler=self.lr_schedulers,
                        step=self.train_step,
                        total_steps=self.total_steps,
                    )
                    self.ckpt_manager.save_check(
                        step=self.train_step,
                        val_score=val_score,
                        pp_enabled=self.parallel_dims.pp_enabled,
                        pp_last_stage=pp_last_stage,
                        pp_master_rank=self.parallel_dims.world_size
                        - self.parallel_dims.world_size / self.parallel_dims.pp,
                    )
            if (
                self.config.train.max_num_steps is not None
                and self.train_step >= self.total_steps
            ):
                break  # break outer epoch loop

        # process the final step
        val_score = self.validate()
        if (
            self.config.train.ckpt.export_safetensors
            and self.parallel_dims.dp_replicate_coord[0] == 0
        ):
            logger.info(
                f"Saving final huggingface checkpoint to {self.config.train.output_dir}..."
            )
            self.export_safetensors(
                output_dir=self.config.train.output_dir,
                rel_path=os.path.join(
                    "safetensors",
                    f"step_{self.train_step}",
                ),
                trainable_only=False,
                is_final=True,
                dtype=util.str2torch_dtype(self.config.train.param_dtype),
            )
        if self.config.train.ckpt.enable_checkpoint:
            logger.info(
                f"Training finished at step {self.train_step}/{self.total_steps}, saving final cosmos checkpoint..."
            )
            self.ckpt_manager.save_checkpoint(
                model=self.model,
                optimizer=self.optimizers,
                scheduler=self.lr_schedulers,
                step=self.train_step,
                total_steps=self.total_steps,
                is_final=True,
            )
            self.ckpt_manager.save_check(
                step=self.train_step,
                val_score=val_score if self.config.validation.enable else -1,
                pp_enabled=self.parallel_dims.pp_enabled,
                pp_last_stage=pp_last_stage,
                pp_master_rank=self.parallel_dims.world_size
                - self.parallel_dims.world_size / self.parallel_dims.pp,
            )
        self.unregister_from_controller()

    @property
    def pp_loss_fn(self):
        # calculate the loss scaling factor
        mini_batch_size = max(self.config.train.train_policy.mini_batch or 1, 1)
        mini_batch_size = min(
            mini_batch_size, self.config.train.train_batch_per_replica
        )
        loss_scaling_factor = (
            mini_batch_size / self.config.train.train_batch_per_replica
        )
        if self.parallel_dims.dp_shard_enabled:
            dp_group = self.parallel_dims.mesh["dp_shard"].get_group()
        else:
            dp_group = None

        if self.parallel_dims.cp_enabled:
            cp_group = self.parallel_dims.mesh["cp"].get_group()
        else:
            cp_group = None

        return torch.compile(
            partial(
                async_safe_ce,
                loss_scaling_factor=loss_scaling_factor,
                dp_group=dp_group
                if self.config.train.train_policy.balance_dp_token
                else None,
                cp_group=cp_group,
            )
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers import AutoConfig
from typing import Any


def pre_hf_models_patch(hf_config: AutoConfig):
    if (
        hf_config.model_type == "internvl_chat"
        and hasattr(hf_config, "llm_config")
        and hf_config.llm_config.model_type == "gpt_oss"
    ):
        hf_config.vision_config.drop_path_rate = 0.0
        print("Set drop_path_rate to 0.0")


def post_hf_models_patch(hf_config: AutoConfig, model: Any):
    if (
        hf_config.model_type == "internvl_chat"
        and hasattr(hf_config, "llm_config")
        and hf_config.llm_config.model_type == "gpt_oss"
    ):
        model.img_context_token_id = 200021
        print("Set img_context_token_id to 200021")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Any, Dict, Optional, Tuple, Union
from pydantic import BaseModel, Field, model_validator


class ChatMessage(BaseModel):
    """
    A chat message item of a conversation.
    """

    role: str = Field(default=None, choices=["system", "user", "assistant", "tool"])

    """
    For text message content,
    ```python
    "What do you see in this video?"
    ```

    For MultiModel message content,
    We support those types of content for multi-model context:
    ```python
    [
        {"type": "text", "text": "What do you see in this video?"},
        {"type": "image", "url": "https://example.com/image.png"},
        {"type": "video", "url": "https://example.com/video.mp4"},
    ]
    ```
    """
    content: str | List[Dict[str, Any]] = ""


ConversationType = List[ChatMessage]


class RLPayload(BaseModel):
    """
    The payload schema of RL sample.
    """

    prompt: Optional[Union[ConversationType, str]] = Field(
        default=None, description="The input prompt for the rollout."
    )

    conversation: Optional[ConversationType] = Field(
        default=None, description="The input conversation for the rollout."
    )

    reference_answer: Optional[str] = Field(
        default=None, description="The reference answer for the rollout."
    )

    weight_version: int = Field(
        default=0, description="The weight version for the rollout."
    )

    # For rollout generation result, we add following fields:
    completions: Optional[List[str]] = Field(
        default=None,
        description="The generated completions for the prompt, In multi-turn conversation, it is a list of last message for each turn.",
    )

    completed_conversations: Optional[List[ConversationType]] = Field(
        default=None,
        description="The original input conversation for the rollout, In multi-turn conversation, it is a list of conversation history for each turn.",
    )

    n_ignore_prefix_tokens: Optional[List[int]] = Field(
        default=None,
        description="The number of prefix tokens to ignore when computing reward.",
    )

    rewards: Optional[List[float]] = Field(
        default=None, description="The reward for each completion."
    )

    advantages: Optional[List[float]] = Field(
        default=None, description="The advantage for each completion."
    )

    valid: Optional[bool] = Field(
        default=True, description="Whether the rollout is valid."
    )

    @model_validator(mode="after")
    def check_params_value(self):
        assert self.prompt or self.conversation, "Must set prompt or conversation"
        return self

    @staticmethod
    def collate_fn(
        batch: List["IdxAndRLPayload"],
    ) -> tuple[List[int], List["RLPayload"]]:
        idx_list = []
        payload_list = []

        for idx, payload in batch:
            idx_list.append(idx)
            payload_list.append(payload)

        return idx_list, payload_list


# When we use iter(dataset), we can get the index of the payload in this way
IdxAndRLPayload = Tuple[int, RLPayload]


class Rollout(BaseModel):
    prompt: Optional[Union[ConversationType, str]] = Field(
        default=None, description="The input prompt for the rollout."
    )

    conversation: Optional[ConversationType] = Field(
        default=None, description="The input conversation for the rollout."
    )

    completion: str = Field(
        default="", description="The generated completion for the rollout."
    )

    completed_conversation: Optional[ConversationType] = Field(
        default=None, description="The generated conversation for the rollout."
    )

    is_end: bool = Field(
        default=False, description="Whether the rollout is the last one."
    )

    reward: float = Field(default=0.0, description="The reward for the rollout.")

    advantage: float = Field(default=0.0, description="The advantage for the rollout.")

    prompt_idx: int = Field(
        default=0, description="The index of the prompt for the rollout."
    )

    n_ignore_prefix_tokens: int = 0

    filter_reward: float = Field(
        default=0.0, description="The filter reward for the rollout."
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script is launched by slurm job.
"""

import argparse
import logging
from typing import List
import subprocess
import time
import os
import sys
from argparse import REMAINDER

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from util import NodeLaunchMetadata


logging.basicConfig(level=logging.INFO)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--type",
        type=str,
        required=True,
        choices=["policy", "rollout"],
    )
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="Path to the config file for the policy or rollout script.",
    )

    parser.add_argument(
        "script",
        nargs="?",  # “?” means 0 or 1 occurrences
        default=None,
        help="A user script which can be provided for custom dataset, reward functions, and model registration.",
    )

    parser.add_argument("script_args", nargs=REMAINDER)

    args = parser.parse_args()

    node_list = os.environ["LOCAL_NODE_LIST"].split(" ")
    COSMOS_CONTROLLER_HOST = os.environ["COSMOS_CONTROLLER_HOST"]
    self_node = os.environ["SLURMD_NODENAME"]
    node_launch_metadata: List[NodeLaunchMetadata] = NodeLaunchMetadata.from_json_list(
        os.environ[f"NODE_LAUNCH_METADATA_{args.type.upper()}"]
    )

    # Either the policy or rollout nodes
    logging.info(f"COSMOS_CONTROLLER_HOST: {COSMOS_CONTROLLER_HOST}")
    logging.info(f"NODE LIST: {node_list}")
    assert self_node in node_list, f"self_node {self_node} not in node_list {node_list}"
    self_node_idx = node_list.index(self_node)
    self_node_launch_metadata = node_launch_metadata[self_node_idx]

    cmds = []
    envs = []
    for replica_launch_metadata in self_node_launch_metadata.colocation:
        rendezvous_node = node_list[replica_launch_metadata.rendezvous_node]
        rendezvous_port = replica_launch_metadata.rendezvous_port
        visible_gpus = replica_launch_metadata.visible_gpus
        nnode = replica_launch_metadata.nnode

        logging.info(
            f"Rendezvous node: {rendezvous_node}, rendezvous port: {rendezvous_port}, visible GPUs: {visible_gpus}"
        )
        env = os.environ.copy()
        env["CUDA_VISIBLE_DEVICES"] = ",".join(str(gpu) for gpu in visible_gpus)
        env["VLLM_DISABLE_COMPILE_CACHE"] = "1"
        replica_launch_script = os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "launcher",
            "launch_replica.sh",
        )
        cmd = [
            replica_launch_script,
            "--type",
            args.type,
            "--rdzv-endpoint",
            f"{rendezvous_node}:{rendezvous_port}",
            "--ngpus",
            str(len(visible_gpus)),
            "--nnodes",
            str(nnode),
        ]
        if args.config:
            cmd += ["--config", args.config]
        if args.script:
            cmd += ["--script", args.script]
        if args.script_args:
            cmd.extend(args.script_args)

        cmds.append(cmd)
        envs.append(env)

    procs = [subprocess.Popen(cmd, env=env) for cmd, env in zip(cmds, envs)]
    # block until every process finishes, and propagate any non-zero exit codes
    exit_code = 0
    while len(procs) > 0:
        for i, p in enumerate(procs):
            try:
                # Check if process has finished without blocking
                if p.poll() is not None:
                    returncode = p.returncode
                    if returncode != 0:
                        # Kill and exit if any process fails
                        for p in procs:
                            try:
                                p.kill()
                            except Exception:
                                pass
                        sys.exit(returncode)
                    # Remove completed process from list
                    procs.remove(p)
            except Exception:
                # Terminate all remaining processes
                for p in procs:
                    try:
                        p.kill()
                    except Exception:
                        pass
                sys.exit(1)
        # Small sleep to prevent busy waiting
        time.sleep(0.1)
    sys.exit(exit_code)  # mimic “all-good” or the first failure



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple
from abc import ABC, abstractmethod

from .schema import OpenAIFunctionCallSchema


class ToolParser(ABC):
    @abstractmethod
    def extract_tool_calls(
        self, text: str
    ) -> Tuple[str, list[OpenAIFunctionCallSchema]]:
        """Extract tool calls from the text.

        Args:
            text (str): The responses text.

        Returns:
            Tuple[str, List[OpenAIFunctionCallSchema]]: Content and extracted tool calls.
        """
        raise NotImplementedError



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# patch trtllm according to trtllm version
import tensorrt_llm

trtllm_version = tensorrt_llm.__version__

if trtllm_version == "1.0.0rc6":
    from cosmos_rl.rollout.trtllm_rollout.patches_for_trtllm import trtllm_1_0_0_rc6  # noqa: F401
else:
    raise NotImplementedError(
        f"[Rollout] Unsupported trtllm version: {trtllm_version}."
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re

from cosmos_rl.utils.logging import logger
from . import ToolParser, OpenAIFunctionCallSchema


class HermesToolParser(ToolParser):
    """Adapted from https://github.com/vllm-project/vllm/blob/v0.9.1/vllm/entrypoints/openai/tool_parsers/hermes_tool_parser.py"""

    tool_call_start_token: str = "<tool_call>"
    tool_call_end_token: str = "</tool_call>"
    tool_call_regex = re.compile(r"<tool_call>(.*?)</tool_call>", re.DOTALL)

    def extract_tool_calls(
        self, text: str
    ) -> tuple[str, list[OpenAIFunctionCallSchema]]:
        if (
            self.tool_call_start_token not in text
            or self.tool_call_end_token not in text
        ):
            return text, []

        matches = self.tool_call_regex.findall(text)
        function_calls = []
        for match in matches:
            try:
                function_call = json.loads(match)
                name, arguments = function_call["name"], function_call["arguments"]
                function_calls.append(
                    OpenAIFunctionCallSchema(name=name, arguments=arguments)
                )
            except Exception as e:
                # Take it easy, due to the generation noise, we may have some invalid tool calls
                logger.error(f"Failed to decode tool call: {e}")

        # remaing text exclude tool call tokens
        content = self.tool_call_regex.sub("", text)

        return content, function_calls
